from .info import update_info, get_info
from .info import INFO_FILEPATH, HOMEPAGE_FILEPATH, CHECKPAGE_FILEPATH, ICON_FILEPATH

from .info import INFO_WIKI_NAME, INFO_RAW_TITLE, INFO_DOKU_URL, INFO_LANG, INFO_ICON_URL